/*
  This file is part of TALER
  Copyright (C) 2014-2026 Taler Systems SA

  TALER is free software; you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free Software
  Foundation; either version 2.1, or (at your option) any later version.

  TALER is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License along with
  TALER; see the file COPYING.LGPL.  If not, see
  <http://www.gnu.org/licenses/>
*/
/**
 * @file src/lib/merchant_api_get-management-instances-INSTANCE.c
 * @brief Implementation of the GET /management/instances/$INSTANCE request
 * @author Christian Grothoff
 */
#include "platform.h"
#include <curl/curl.h>
#include <jansson.h>
#include <microhttpd.h> /* just for HTTP status codes */
#include <gnunet/gnunet_util_lib.h>
#include <gnunet/gnunet_curl_lib.h>
#include <taler/merchant/get-management-instances-INSTANCE.h>
#include "merchant_api_curl_defaults.h"
#include <taler/taler_json_lib.h>


/**
 * Maximum number of accounts permitted.
 */
#define MAX_ACCOUNTS 1024


/**
 * Handle for a GET /management/instances/$INSTANCE operation.
 */
struct TALER_MERCHANT_GetManagementInstanceHandle
{
  /**
   * Base URL of the merchant backend.
   */
  char *base_url;

  /**
   * Instance identifier to query.
   */
  char *instance_id;

  /**
   * The full URL for this request.
   */
  char *url;

  /**
   * Handle for the request.
   */
  struct GNUNET_CURL_Job *job;

  /**
   * Function to call with the result.
   */
  TALER_MERCHANT_GetManagementInstanceCallback cb;

  /**
   * Closure for @a cb.
   */
  TALER_MERCHANT_GET_MANAGEMENT_INSTANCE_RESULT_CLOSURE *cb_cls;

  /**
   * Reference to the execution context.
   */
  struct GNUNET_CURL_Context *ctx;
};


/**
 * Parse account information from @a ia.
 *
 * @param ia JSON array (or NULL!) with account data
 * @param[in,out] igr partially filled response
 * @param gmi operation handle
 * @return #GNUNET_OK on success
 */
static enum GNUNET_GenericReturnValue
parse_accounts (const json_t *ia,
                struct TALER_MERCHANT_GetManagementInstanceResponse *igr,
                struct TALER_MERCHANT_GetManagementInstanceHandle *gmi)
{
  unsigned int accounts_len = (unsigned int) json_array_size (ia);

  if ( (json_array_size (ia) != (size_t) accounts_len) ||
       (accounts_len > MAX_ACCOUNTS) )
  {
    GNUNET_break (0);
    return GNUNET_SYSERR;
  }
  {
    struct TALER_MERCHANT_ManagementAccountEntry accounts[
      GNUNET_NZL (accounts_len)];
    size_t index;
    json_t *value;

    memset (accounts,
            0,
            sizeof (accounts));
    json_array_foreach (ia, index, value) {
      struct TALER_MERCHANT_ManagementAccountEntry *ae =
        &accounts[index];
      struct GNUNET_JSON_Specification spec[] = {
        TALER_JSON_spec_full_payto_uri ("payto_uri",
                                        &ae->payto_uri),
        GNUNET_JSON_spec_mark_optional (
          TALER_JSON_spec_web_url (
            "credit_facade_url",
            &ae->credit_facade_url),
          NULL),
        GNUNET_JSON_spec_fixed_auto ("h_wire",
                                     &ae->h_wire),
        GNUNET_JSON_spec_fixed_auto ("salt",
                                     &ae->salt),
        GNUNET_JSON_spec_bool ("active",
                               &ae->active),
        GNUNET_JSON_spec_end ()
      };

      if (GNUNET_OK !=
          GNUNET_JSON_parse (value,
                             spec,
                             NULL, NULL))
      {
        GNUNET_break_op (0);
        return GNUNET_SYSERR;
      }
    }
    igr->details.ok.accounts_length = accounts_len;
    igr->details.ok.accounts = accounts;
    gmi->cb (gmi->cb_cls,
             igr);
    gmi->cb = NULL;
  }
  return GNUNET_OK;
}


/**
 * Function called when we're done processing the
 * HTTP GET /management/instances/$INSTANCE request.
 *
 * @param cls the `struct TALER_MERCHANT_GetManagementInstanceHandle`
 * @param response_code HTTP response code, 0 on error
 * @param response response body, NULL if not in JSON
 */
static void
handle_get_instance_finished (void *cls,
                              long response_code,
                              const void *response)
{
  struct TALER_MERCHANT_GetManagementInstanceHandle *gmi = cls;
  const json_t *json = response;
  struct TALER_MERCHANT_GetManagementInstanceResponse igr = {
    .hr.http_status = (unsigned int) response_code,
    .hr.reply = json
  };

  gmi->job = NULL;
  GNUNET_log (GNUNET_ERROR_TYPE_DEBUG,
              "Got /management/instances/$INSTANCE response with status code %u\n",
              (unsigned int) response_code);
  switch (response_code)
  {
  case MHD_HTTP_OK:
    {
      const json_t *address;
      const json_t *jurisdiction;
      const json_t *accounts;
      struct GNUNET_JSON_Specification spec[] = {
        GNUNET_JSON_spec_string (
          "name",
          &igr.details.ok.details.name),
        GNUNET_JSON_spec_fixed_auto (
          "merchant_pub",
          &igr.details.ok.details.merchant_pub),
        GNUNET_JSON_spec_object_const (
          "address",
          &address),
        GNUNET_JSON_spec_object_const (
          "jurisdiction",
          &jurisdiction),
        GNUNET_JSON_spec_bool (
          "use_stefan",
          &igr.details.ok.details.use_stefan),
        GNUNET_JSON_spec_relative_time (
          "default_wire_transfer_delay",
          &igr.details.ok.details.default_wire_transfer_delay),
        GNUNET_JSON_spec_relative_time (
          "default_pay_delay",
          &igr.details.ok.details.default_pay_delay),
        GNUNET_JSON_spec_mark_optional (
          GNUNET_JSON_spec_relative_time (
            "default_refund_delay",
            &igr.details.ok.details.default_refund_delay),
          NULL),
        GNUNET_JSON_spec_mark_optional (
          GNUNET_JSON_spec_time_rounder_interval (
            "default_wire_transfer_rounding_interval",
            &igr.details.ok.details.default_wire_transfer_rounding_interval),
          NULL),
        GNUNET_JSON_spec_mark_optional (
          GNUNET_JSON_spec_string (
            "email",
            &igr.details.ok.details.email),
          NULL),
        GNUNET_JSON_spec_mark_optional (
          GNUNET_JSON_spec_bool (
            "email_validated",
            &igr.details.ok.details.email_validated),
          NULL),
        GNUNET_JSON_spec_mark_optional (
          GNUNET_JSON_spec_string (
            "phone_number",
            &igr.details.ok.details.phone_number),
          NULL),
        GNUNET_JSON_spec_mark_optional (
          GNUNET_JSON_spec_bool (
            "phone_validated",
            &igr.details.ok.details.phone_validated),
          NULL),
        GNUNET_JSON_spec_mark_optional (
          TALER_JSON_spec_web_url (
            "website",
            &igr.details.ok.details.website),
          NULL),
        GNUNET_JSON_spec_mark_optional (
          GNUNET_JSON_spec_string (
            "logo",
            &igr.details.ok.details.logo),
          NULL),
        GNUNET_JSON_spec_array_const (
          "accounts",
          &accounts),
        GNUNET_JSON_spec_end ()
      };

      if (GNUNET_OK !=
          GNUNET_JSON_parse (json,
                             spec,
                             NULL, NULL))
      {
        GNUNET_break_op (0);
        igr.hr.http_status = 0;
        igr.hr.ec = TALER_EC_GENERIC_INVALID_RESPONSE;
        break;
      }
      igr.details.ok.details.address = address;
      igr.details.ok.details.jurisdiction = jurisdiction;
      if (GNUNET_OK ==
          parse_accounts (accounts,
                          &igr,
                          gmi))
      {
        TALER_MERCHANT_get_management_instance_cancel (gmi);
        return;
      }
      igr.hr.http_status = 0;
      igr.hr.ec = TALER_EC_GENERIC_INVALID_RESPONSE;
      break;
    }
  case MHD_HTTP_UNAUTHORIZED:
    igr.hr.ec = TALER_JSON_get_error_code (json);
    igr.hr.hint = TALER_JSON_get_error_hint (json);
    break;
  case MHD_HTTP_NOT_FOUND:
    igr.hr.ec = TALER_JSON_get_error_code (json);
    igr.hr.hint = TALER_JSON_get_error_hint (json);
    break;
  default:
    igr.hr.ec = TALER_JSON_get_error_code (json);
    igr.hr.hint = TALER_JSON_get_error_hint (json);
    GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
                "Unexpected response code %u/%d\n",
                (unsigned int) response_code,
                (int) igr.hr.ec);
    break;
  }
  gmi->cb (gmi->cb_cls,
           &igr);
  TALER_MERCHANT_get_management_instance_cancel (gmi);
}


struct TALER_MERCHANT_GetManagementInstanceHandle *
TALER_MERCHANT_get_management_instance_create (
  struct GNUNET_CURL_Context *ctx,
  const char *url,
  const char *instance_id)
{
  struct TALER_MERCHANT_GetManagementInstanceHandle *gmi;

  gmi = GNUNET_new (struct TALER_MERCHANT_GetManagementInstanceHandle);
  gmi->ctx = ctx;
  gmi->base_url = GNUNET_strdup (url);
  if (NULL != instance_id)
    gmi->instance_id = GNUNET_strdup (instance_id);
  return gmi;
}


enum TALER_ErrorCode
TALER_MERCHANT_get_management_instance_start (
  struct TALER_MERCHANT_GetManagementInstanceHandle *gmi,
  TALER_MERCHANT_GetManagementInstanceCallback cb,
  TALER_MERCHANT_GET_MANAGEMENT_INSTANCE_RESULT_CLOSURE *cb_cls)
{
  CURL *eh;

  gmi->cb = cb;
  gmi->cb_cls = cb_cls;
  if (NULL != gmi->instance_id)
  {
    char *path;

    GNUNET_asprintf (&path,
                     "instances/%s/private",
                     gmi->instance_id);
    gmi->url = TALER_url_join (gmi->base_url,
                               path,
                               NULL);
    GNUNET_free (path);
  }
  else
  {
    gmi->url = TALER_url_join (gmi->base_url,
                               "private",
                               NULL);
  }
  if (NULL == gmi->url)
    return TALER_EC_GENERIC_CONFIGURATION_INVALID;
  eh = TALER_MERCHANT_curl_easy_get_ (gmi->url);
  if (NULL == eh)
    return TALER_EC_GENERIC_CONFIGURATION_INVALID;
  gmi->job = GNUNET_CURL_job_add (gmi->ctx,
                                  eh,
                                  &handle_get_instance_finished,
                                  gmi);
  if (NULL == gmi->job)
    return TALER_EC_GENERIC_INTERNAL_INVARIANT_FAILURE;
  return TALER_EC_NONE;
}


void
TALER_MERCHANT_get_management_instance_cancel (
  struct TALER_MERCHANT_GetManagementInstanceHandle *gmi)
{
  if (NULL != gmi->job)
  {
    GNUNET_CURL_job_cancel (gmi->job);
    gmi->job = NULL;
  }
  GNUNET_free (gmi->url);
  GNUNET_free (gmi->instance_id);
  GNUNET_free (gmi->base_url);
  GNUNET_free (gmi);
}


/* end of merchant_api_get-management-instances-INSTANCE.c */
